// deltarule.cpp
//
// Compile with: g++ deltarule.cpp -o deltarule
//
// Created by: Alejandro Gutierrez
// Created Date: 2/15/2008
// Last Modified Date: 12/18/2008
// University of South Florida
// CIS 6930-013 Machine Learning

#include <iostream>
#include <fstream>
#include <sstream>
#include <string>
#include <vector>
#include <ctime>
#include <iomanip>
#include <math.h>

using namespace std;

// User Defined Datatypes
typedef vector< double > TrainingExample;
typedef vector< double > Weights;

// Function prototypes
void parse_command_line(int,char**, string&, double&, bool&, bool&, double&);
void exit_with_help();
vector< TrainingExample > LoadTrainingExamples(const string&);
void GradientDescent(const vector< TrainingExample >&, const double&, const bool&, const bool&, const double&);
double ComputeOutput(const TrainingExample&, const Weights&);

int main(int argc, char* argv[]){

	string filename;
	double LearningRate;
	bool DecreaseLR;
	bool Batch;
	double Epsilon;
	vector< TrainingExample > TrainingExamples;
	time_t StartTime;
	time_t FinalTime;

	parse_command_line(argc,argv,filename,LearningRate,DecreaseLR,Batch,Epsilon);

	//get start time
	StartTime = time(NULL);

		
	//load the training examples
	TrainingExamples = LoadTrainingExamples(filename);

	// show selected options
	cout << endl << "OPTIONS" << endl;
	cout << "Learning Rate: " << LearningRate << endl;
	if (DecreaseLR) cout << "Decrease LR: Yes" << endl;
	if (Batch) cout << "Batch Processing: Yes" << endl;
	
	//Start training
	GradientDescent(TrainingExamples,LearningRate,DecreaseLR,Batch,Epsilon);

	//display final time
	FinalTime = time(NULL);
	cout << endl << endl << "Total Execution Time: " << (FinalTime  - StartTime) << " seconds" << endl << endl;

	return 0;

}

void exit_with_help() {
	
	cout <<	"Usage: deltarule [options] <trainingfile.txt>" << endl;
	cout << "options:" << endl;
	cout << "-r  <number> Learning rate. Default 0.1 (must greater than 0)" << endl;
	cout << "-d  Decreases the learning rate on each iteration" << endl;  
	cout << "-b  Batch processing. (Gradient Descent)" << endl; 
	cout << "-e  <number> Epsilon. Default 1.0e-10 (must greater than 0)" << endl; 
	
	exit(1);
}


void parse_command_line(int argc,char* argv[], string &TrainingFile, double &LearningRate, bool &DecreaseLR, bool &Batch, double &Epsilon) {

	int i;

	// default values
	LearningRate = 0.1;
	DecreaseLR = false;
	Batch = false;
	Epsilon =  0.0000000001;

	// parse options
	for(i=1;i<argc;i++)
	{
		if(argv[i][0] != '-') break;
		if(++i>=argc)
			exit_with_help();
		switch(argv[i-1][1])
		{
			case 'e':
				Epsilon = atof(argv[i]);
				if (Epsilon < 0.0)
					exit_with_help();
				break;
			case 'r':
				LearningRate = atof(argv[i]);
				if (LearningRate < 0.0)
					exit_with_help();
				break;
			case 'd':
				DecreaseLR = true;
				i--;
				break;
			case 'b':
				Batch = true;
				i--;
				break;
			default:
				cerr << "unknown option" << endl;
				exit_with_help();
		}
	}

	// determine filename
	if(i>=argc)
		exit_with_help();

	TrainingFile =  argv[i];

	return;

}

vector< TrainingExample > LoadTrainingExamples(const string &filename){

	ifstream TraningFile;
	istringstream buffer;
	TrainingExample trainingInstanceTemp;
	vector < TrainingExample > TrainingExamples;
	string strTemp;
	double fTemp; 

	TraningFile.open(filename.c_str());
	
	if (TraningFile.is_open()) {

		cout << flush <<  "Loading Training File: " << filename <<  "...   ";

		// bias, X0, X1, Target
		trainingInstanceTemp.resize(4); 

		//read the column names row
		getline(TraningFile,strTemp);

		while (!TraningFile.eof()) {

			getline(TraningFile,strTemp);
			buffer.str(strTemp);

			if (!strTemp.empty()) {

				trainingInstanceTemp[0] = 1.0;

				buffer >> fTemp;
				trainingInstanceTemp[3] = fTemp; // output

				buffer >> fTemp;
				trainingInstanceTemp[1] = fTemp; // x1

				buffer >> fTemp;
				trainingInstanceTemp[2] = fTemp; // x2
				
				TrainingExamples.push_back(trainingInstanceTemp);
			}

			buffer.clear(); 
		}

		TraningFile.close();

		cout <<  "Done." << endl;

	}
	else {
		cerr << flush << "Traning File not found." << endl;
	}

	return TrainingExamples;

}


void GradientDescent(const vector< TrainingExample > &TrainingExamples, const double &LearningRate, const bool &DecreaseLR, const bool &Batch, const double &Epsilon){

	Weights CurrentWeights(3);
	Weights DeltaWeights(3);
	TrainingExample x;
	double LR;
	double output;
	double target;
	double error = 0.0;
	double prev_error = 0.0;
	double max_weight = 0.0;
	unsigned long epochs = 0;
	unsigned long WeightUpdates = 0;

	cout << endl << "Training...    Error:       ";

	// set the seed
	srand ( (unsigned int) time(NULL) );

	LR = LearningRate;
	
	// Initialize each Wi to some small random value
	for (int i=0; i < (int) CurrentWeights.size(); i++ ){
		CurrentWeights[i] = 1.0/((rand()%1000)+1.0);
	}

	// Until the termination condition is met
	do {

		epochs++;
		
		//decrease LR based on iterations
		if (DecreaseLR) {
			LR = LearningRate/((double)(epochs));
		}

		if (Batch) {
			// Initialize each DeltaWi to zero
			for (int i=0; i < (int) DeltaWeights.size(); i++ ){
				DeltaWeights[i] = 0.0;
			}
		}

		// For each <x,t> in training_examples, Do
		for (int i=0; i < (int) TrainingExamples.size(); i++ ){

			x = TrainingExamples[i];
			target = x.back();

			// input the instance x to the unit and compute the output
			output = ComputeOutput(x,CurrentWeights);

			if (Batch) {
				// for each linear unit weight Wi, Do   deltaWi = deltaWi + n(t-o)Xi
				// where target = t, LR = n, and output = o
				for (int j=0; j < (int) DeltaWeights.size(); j++ ){
					DeltaWeights[j] = DeltaWeights[j] + LR*(target-output)*x[j];
				}

				// error for gradient descent
				error += pow((target-output),2.0);
			}
			else {
				//stochastic approximation
				for (int j=0; j < (int) CurrentWeights.size(); j++ ){
					CurrentWeights[j] = CurrentWeights[j] + LR*(target-output)*x[j];
					WeightUpdates++;
				}
				
				// error for stochastic approach
				error = pow((target-output),2.0)/2.0;

				// if error still the same as 10000 epochs ago
				if (fabs(error - prev_error) < Epsilon){
					break;
				}

			}
		}

		if (Batch) {
			// for each linear unit weight Wi, Do   Wi = Wi + deltaWi
			for (int i=0; i < (int) CurrentWeights.size(); i++ ){
				CurrentWeights[i] = CurrentWeights[i] + DeltaWeights[i];
				WeightUpdates++;
			}

			// E(w) = 1/2 Sum(td-Od,d in D)^2
			error = error/2.0;

			// if error still the same as 10000 epochs ago
			if (fabs(error - prev_error) < Epsilon){
				break;
			}
	
		}


		// report error each 1000 steps
		if (epochs%1000==0) {
				cout << "\b\b\b\b\b\b" << setiosflags(ios::fixed) << setprecision(4) << error ; 
				prev_error = error;
		}

		//use this section to get points to plot
		/*if (epochs < 20 || epochs == 50 || epochs == 100 || epochs == 250 || epochs == 500 || epochs == 1000 || epochs == 10000 || epochs == 20000 || epochs == 50000 || epochs == 100000 || epochs%1000000==0){
			cout << endl << epochs << "," << setiosflags(ios::fixed) << setprecision(4) << error;
		}*/
		
		if (error > 9.99e99) {
			cout << endl << endl << "FATAL ERROR: Please use a smaller learning rate." << endl;
			return;
		}
	
	} while (error > Epsilon);
			

	//normalize weights for display
	for (int i=0; i < (int) CurrentWeights.size(); i++ ){
		if (max_weight < fabs(CurrentWeights[i])) {
			max_weight = fabs(CurrentWeights[i]);
		}
	}

	for (int i=0; i < (int) CurrentWeights.size(); i++ ){
		CurrentWeights[i] = CurrentWeights[i]/max_weight;
	}
	
	// print weights
	cout << endl << endl << "Final Weights:" << endl;
	for (int i=0; i < (int) CurrentWeights.size(); i++ ){
		cout << "W" << i << ": " << setiosflags(ios::fixed) << setprecision(3) << CurrentWeights[i] << "    ";
	}

	cout << endl << endl << "Number of Epochs: " << epochs << endl;
	cout << "Number of Weight Updates: " << WeightUpdates;

	return;
}


double ComputeOutput(const TrainingExample &x, const Weights &w){

	double output = 0.0;

	// don't use x's last element as it corresponds to the target value
	for (int i=0; i < (int) x.size() - 1; i++ ){
		output += x[i]*w[i]; 
	}

	return output;

}